"""Help messages for the CLI."""

from . import console
from .find_subcommands import find_subcommands, warn_if_completion_not_installed

STYLE_OPTION = "bold cyan"
STYLE_SWITCH = "bold green"
STYLE_NEGATIVE_OPTION = "bold magenta"
STYLE_NEGATIVE_SWITCH = "bold red"

STYLE_OPTIONS_TABLE_SHOW_LINES = False
STYLE_OPTIONS_TABLE_LEADING = 0
STYLE_OPTIONS_TABLE_PAD_EDGE = False
STYLE_OPTIONS_TABLE_PADDING = (0, 1)
STYLE_OPTIONS_TABLE_BOX = ""
STYLE_OPTIONS_TABLE_ROW_STYLES = None
STYLE_OPTIONS_TABLE_BORDER_STYLE = None


def print_help():
    """Print help for the main CLI."""

    from rich.panel import Panel  # noqa: PLC0415  lazy loading of rich
    from rich.table import Table  # noqa: PLC0415  lazy loading of rich

    console.print("[yellow]Usage:[/yellow] [bold]slices [OPTIONS] COMMAND [ARGS]...[/bold]")
    console.print()
    console.print("Slices CLI")
    console.print()

    options_table = Table(highlight=False, show_header=False, expand=True, box=None)
    options_table.add_column(style=STYLE_OPTION, no_wrap=True)
    options_table.add_column(style=STYLE_SWITCH, no_wrap=True)
    options_table.add_column(justify="left", no_wrap=False, ratio=10)
    options_table.add_row("--help", "-h", "Show this message and exit.")
    options_table.add_row("--version", "", "Show CLI version.")
    options_table.add_row("--no-upgrade-check", "", "Do not check for newer version.")

    envvar_table = Table(highlight=False, show_header=False, expand=True, box=None)
    envvar_table.add_column(style=STYLE_OPTION, no_wrap=True)
    envvar_table.add_column(justify="left", no_wrap=False, ratio=10)
    envvar_table.add_row("SLICES_CUSTOM_CONFIG", "Override core config with file content.")
    envvar_table.add_row("SLICES_BI_CUSTOM_CONFIG", "Override BI config with file content.")
    envvar_table.add_row("SLICES_DEBUG", "Debug output level. 0, 1, 2 or 3. Default: 0")

    options_panel = Panel(
        options_table,
        title="Options",
        title_align="left",
        border_style="bright_black",
        expand=True,
    )
    envvars_panel = Panel(
        envvar_table,
        title="Environment Variables (advanced)",
        title_align="left",
        border_style="bright_black",
        expand=True,
    )
    # commands_panel = Panel(
    #     f"[{STYLE_OPTION}]clear-cache[/{STYLE_OPTION}]   Remove all cache files.",
    #     title="Commands",
    #     title_align="left",
    #     border_style="bright_black",
    #     expand=True,
    # )
    console.print(options_panel)
    console.print(envvars_panel)
    # console.print(commands_panel)
    subcommands = find_subcommands()
    warn_if_completion_not_installed(subcommands)

    subcommands_table = Table(highlight=False, show_header=False, expand=True, box=None)
    # max_cmd_len = max([len(subcommand.name) for subcommand in subcommands], default=0)
    # add_column(max_width=max_cmd_len
    subcommands_table.add_column(style="bold cyan", no_wrap=True)
    subcommands_table.add_column("Description", justify="left", no_wrap=False, ratio=10)
    subcommands_table.add_row("clear-cache", "Remove all cache files")
    subcommands_table.add_row("completion", "Show or install shell completions")
    for subcommand in subcommands:
        alias_str = ""
        for alias in subcommand.visible_aliases:
            alias_str += f" | {alias}"
        subcommands_table.add_row(subcommand.name + alias_str, subcommand.short_help)

    subcommands_panel = Panel(
        subcommands_table,
        # title="Commands with subcommands",
        title="Commands",
        title_align="left",
        border_style="bright_black",
        expand=True,
    )

    console.print(subcommands_panel)


def print_clear_cache_help():
    """Print help for clear-cache command."""

    from rich.panel import Panel  # noqa: PLC0415  lazy loading of rich

    console.print("[yellow]Usage:[/yellow] [bold]slices clear-cache [OPTIONS][/bold]")
    console.print()
    console.print("Remove all cache files.")
    console.print()

    options_panel = Panel(
        (
            f"[{STYLE_OPTION}]--help[/{STYLE_OPTION}]"
            f"                [{STYLE_SWITCH}]-h[/{STYLE_SWITCH}]"
            "        Show this message and exit."
        ),
        title="Options",
        title_align="left",
        border_style="bright_black",
        expand=True,
    )

    console.print(options_panel)


def print_completion_help():
    """Print help for completion command."""

    from rich.panel import Panel  # noqa: PLC0415  lazy loading of rich

    console.print("[yellow]Usage:[/yellow]")
    console.print("   [bold]slices completion install [OPTIONS][/bold]")
    console.print("   Install completion for the current shell.")
    console.print()
    console.print("   [bold]slices completion show [OPTIONS][/bold]")
    console.print("   Show completion for the current shell, to copy it or customize the installation.")
    console.print()

    options_panel = Panel(
        (
            f"[{STYLE_OPTION}]--help[/{STYLE_OPTION}]"
            f"                [{STYLE_SWITCH}]-h[/{STYLE_SWITCH}]"
            "        Show this message and exit."
        ),
        title="Options",
        title_align="left",
        border_style="bright_black",
        expand=True,
    )

    console.print(options_panel)



"""Main Slices CLI.

This CLI mostly calls others CLIs.
"""

import os
from pathlib import Path

from slices_cli.console import DummyConsole, LazyRichConsole

console = LazyRichConsole(log_path=False)
err_console = LazyRichConsole(stderr=True)

DEBUG_LEVEL_INFO = 1
DEBUG_LEVEL_DEBUG = 2
DEBUG_LEVEL_TRACE = 3

try:
    DEBUG_LEVEL = int(os.environ.get("SLICES_DEBUG", "0"))
except ValueError:
    DEBUG_LEVEL = 0
    err_console.print("[red bold] :x: ignoring SLICES_DEBUG environment variable: not an integer", highlight=False)

console_info = LazyRichConsole(log_path=False) if DEBUG_LEVEL >= DEBUG_LEVEL_INFO else DummyConsole()
console_debug = LazyRichConsole(log_path=False) if DEBUG_LEVEL >= DEBUG_LEVEL_DEBUG else DummyConsole()
console_trace = LazyRichConsole(log_path=False) if DEBUG_LEVEL >= DEBUG_LEVEL_TRACE else DummyConsole()

CONFIG_DIR = ".slices"
config_dir_location = Path.home() / CONFIG_DIR
cache_dir_location = config_dir_location / "cache"



"""Shell completion."""

# The code in this file was partially copied from:
#    https://github.com/fastapi/typer/blob/6f2ec1d00ec76aa1d2b8743121dacdd48f4acbaa/typer/_completion_shared.py
#
# Which has this MIT License:
#
#    The MIT License (MIT)
#
#    Copyright (c) 2019 Sebastián Ramírez
#
#    Permission is hereby granted, free of charge, to any person obtaining a copy
#    of this software and associated documentation files (the "Software"), to deal
#    in the Software without restriction, including without limitation the rights
#    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the Software is
#    furnished to do so, subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be included in
#    all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#    THE SOFTWARE.

import sys
from pathlib import Path
from subprocess import TimeoutExpired, run

import shellingham

from . import console, console_trace, err_console
from .cache import save_cached_subcommand_info
from .find_subcommands import SubCommand, find_subcommands
from .options import OPTIONS_WITH_ARGS, OPTIONS_WITH_ARGS__HELP, OPTIONS_WITHOUT_ARGS, OPTIONS_WITHOUT_ARGS__HELP


def _bash_completion_script() -> str:
    """Hack together a bash completion script that calls subcommands directly to complete them."""

    subcommand_completions = ""
    no_subcommand_compreply = ["pubkey", "auth", "project", "experiment", "clear-cache", "completion"]
    no_subcommand_compreply.extend(OPTIONS_WITHOUT_ARGS)
    no_subcommand_compreply.extend(OPTIONS_WITH_ARGS)

    subcommand_completions += (
        "\n\n"
        "    # Completions for subcommand 'completion'\n"
        """    if [[ -n "$SUBCOMMAND" ]] && ( [ "$SUBCOMMAND" = 'completion' ] )
    then
        if [[ 2 -eq "$COMP_CWORD" ]]
        then
           local IFS=$' \\t\\n'
           COMPREPLY=($(compgen -W 'install show --help' -- "${COMP_WORDS[-1]}"))
        else
           COMPREPLY=()
        fi
        return 0
    fi
    """
    )

    subcommands = find_subcommands()
    for subcommand in subcommands:
        aliases_match = ""
        for alias in subcommand.visible_aliases + subcommand.hidden_aliases:
            aliases_match += f" || [ \"$SUBCOMMAND\" = '{alias}' ]"

        no_subcommand_compreply.extend(subcommand.visible_aliases)

        if not subcommand.is_core:
            no_subcommand_compreply.append(subcommand.name)
        if subcommand.tab_completion_supported and not subcommand.is_core:
            name = subcommand.name

            # this assumes the completions are registered with: complete -F...
            subcommand_completions += (
                "\n\n"
                f"    # Completions for subcommand {name}\n"
                f"""    if [[ -n "$SUBCOMMAND" ]] && ( [ "$SUBCOMMAND" = '{name}' ]{aliases_match} )"""
                """
    then
        # Use the completion function already registered
        completion_func=$(complete -p slices-"""
                + name
                + """ | awk '{for(i=1;i<=NF;i++) if($i=="-F") print $(i+1)}')
        "$completion_func" '"""
                + str(subcommand.executable)
                + """'
        return 0
    fi
            """
            )
        if not subcommand.tab_completion_supported:
            name = subcommand.name
            subcommand_completions += (
                "\n\n"
                f"    # Completions for subcommand {name}\n"
                f"""    if [[ -n "$SUBCOMMAND" ]] && ( [ "$SUBCOMMAND" = '{name}' ]{aliases_match} )"""
                """
    then
        # No completion support for this subcommand
        COMPREPLY=()
        return 0
    fi
            """
            )

    return (
        """_slices_completion() {
    local IFS=$'
'
    # Find the subcommand
    SUBCOMMAND=""
    SUBCOMMAND_INDEX="-1"
    for index in "${!COMP_WORDS[@]}"; do
        if [[ "$index" -eq 0 ]]; then
            continue
        fi
        word="${COMP_WORDS[index]}"
        if [[ ! "$word" =~ ^- ]]; then
            SUBCOMMAND="$word"
            SUBCOMMAND_INDEX="$index"
            break
        fi
    done

    if [[ -z "$SUBCOMMAND" ]] || [[ "$SUBCOMMAND_INDEX" -eq "$COMP_CWORD" ]]
    then
        local IFS=$' \\t\\n'
        COMPREPLY=($(compgen -W '"""
        + (" ".join(no_subcommand_compreply))
        + """' -- "${COMP_WORDS[-1]}"))
        return 0
    fi
    """
        + subcommand_completions
        + """

    # For all other cases, use completions for slices-core (these handle the main slices CLI completions as well.)
    COMPREPLY=( $( env COMP_WORDS="${COMP_WORDS[*]}" \
                   COMP_CWORD=$COMP_CWORD \
                   _SLICES_CORE_COMPLETE=complete_bash $1-core ) )
    return 0
}

complete -o default -F _slices_completion slices"""
    )


def _zsh_completion_script() -> str:
    """Hack together a zsh completion script that calls subcommands directly to complete them."""

    subcommand_completions = ""
    no_subcommand_compreply = {
        "pubkey": "Public key",
        "auth": "Authentication",
        "project": "Projects",
        "experiment": "Experiments",
        "clear-cache": "Remove all cache files.",
        "completion": "Install or show shell completion scripts.",
    }
    no_subcommand_compreply.update(OPTIONS_WITHOUT_ARGS__HELP)
    no_subcommand_compreply.update(OPTIONS_WITH_ARGS__HELP)

    subcommand_completions += (
        "\n\n"
        "    # Completions for subcommand 'completion'\n"
        """    if [[ -n "$subcommand" ]] && ( [ "$subcommand" = 'completion' ] ); then
        subcommands=(
            'install:Install Completion scripts'
            'show:Show Completion scripts'
        )
        _describe 'command' subcommands
        return $?
    fi
    """
    )

    subcommands = find_subcommands()
    for subcommand in subcommands:
        aliases_match = ""
        for alias in subcommand.visible_aliases + subcommand.hidden_aliases:
            aliases_match += f" || [ \"$subcommand\" = '{alias}' ]"

        for alias in subcommand.visible_aliases:
            no_subcommand_compreply[alias] = subcommand.short_help + " (alias)"

        if not subcommand.is_core:
            no_subcommand_compreply[subcommand.name] = subcommand.short_help
        if subcommand.tab_completion_supported and not subcommand.is_core:
            name = subcommand.name

            # this assumes the completions are registered with: complete -F...
            subcommand_completions += (
                "\n\n"
                f"    # Completions for subcommand {name}\n"
                f"""    if [[ -n "$subcommand" ]] && ( [ "$subcommand" = '{name}' ]{aliases_match} ); then"""
                """
        # Use the completion file registered in zsh dir
        _dispatch slices-"""
                + name
                + """bi slices-"""
                + name
                + """
        return $?
    fi
            """
            )
        if not subcommand.tab_completion_supported:
            name = subcommand.name
            subcommand_completions += (
                "\n\n"
                f"    # Completions for subcommand {name}\n"
                f"""    if [[ -n "$subcommand" ]] && ( [ "$subcommand" = '{name}' ]{aliases_match} ); then"""
                """
        # No completion support for this subcommand
        return 0
    fi
            """
            )

    return (
        """#compdef slices

_slices_completion() {
    # Set up Zsh completion system variables
    local -a subcommands
    local curcontext="$curcontext"
    typeset -A opt_args

    local subcommand=""
    local subcommand_index="-1"
    for i in {2..$CURRENT}; do
        local word="${words[i]}"
        # If the word doesn't start with a '-', it's likely the subcommand
        if [[ ! "$word" =~ ^- ]]; then
            subcommand="$word"
            subcommand_index="$i"
            break
        fi
    done

    if [[ -z "$subcommand" ]] || [[ "$subcommand_index" -eq "$CURRENT" ]]; then
        subcommands=(
        """
        + ("\n".join([f"'{k}:{v}'" for k, v in no_subcommand_compreply.items()]))
        + """
        )
        _describe 'command' subcommands
        return 0
    fi

    """
        + subcommand_completions
        + """

    _dispatch slices-core slices-core

    return 0
}

compdef _slices_completion slices
"""
    )


def install_bash(*, prog_name: str) -> Path:
    """Install the bash completion script and source it in ~/.bashrc."""

    # Ref: https://github.com/scop/bash-completion#faq
    # It seems bash-completion is the official completion system for bash:
    # Ref: https://www.gnu.org/software/bash/manual/html_node/A-Programmable-Completion-Example.html
    # But installing in the locations from the docs doesn't seem to have effect
    completion_path = Path.home() / ".bash_completions" / f"{prog_name}.sh"
    rc_path = Path.home() / ".bashrc"
    rc_path.parent.mkdir(parents=True, exist_ok=True)
    rc_content = ""
    if rc_path.is_file():
        rc_content = rc_path.read_text()
    completion_init_lines = [f"source '{completion_path}'"]
    for line in completion_init_lines:
        if line not in rc_content:  # pragma: no cover
            rc_content += f"\n{line}"
    rc_content += "\n"
    rc_path.write_text(rc_content)
    # Install completion
    completion_path.parent.mkdir(parents=True, exist_ok=True)
    script_content = _bash_completion_script()
    completion_path.write_text(script_content)
    return completion_path


def install_zsh(*, prog_name: str) -> Path:
    """Install the zsh completion script."""

    # Setup Zsh and load ~/.zfunc
    zshrc_path = Path.home() / ".zshrc"
    zshrc_path.parent.mkdir(parents=True, exist_ok=True)
    zshrc_content = ""
    if zshrc_path.is_file():
        zshrc_content = zshrc_path.read_text()
    completion_line = "fpath+=~/.zfunc; autoload -Uz compinit; compinit"
    if completion_line not in zshrc_content:
        zshrc_content += f"\n{completion_line}\n"
    style_line = "zstyle ':completion:*' menu select"
    # TODO: consider setting the style only for the current program
    # style_line = f"zstyle ':completion:*:*:{prog_name}:*' menu select"
    # Install zstyle completion config only if the user doesn't have a customization
    if "zstyle" not in zshrc_content:
        zshrc_content += f"\n{style_line}\n"
    zshrc_content = f"{zshrc_content.strip()}\n"
    zshrc_path.write_text(zshrc_content)
    # Install completion under ~/.zfunc/
    path_obj = Path.home() / f".zfunc/_{prog_name}"
    path_obj.parent.mkdir(parents=True, exist_ok=True)
    script_content = _zsh_completion_script()
    path_obj.write_text(script_content)
    return path_obj


def _fish_completion_script():
    """Generate the fish completion script."""

    console_trace.print("[bright_black]Generating fish completion script...")

    raise NotImplementedError("Not yet implemented for fish")


def install_fish(*, prog_name: str) -> Path:
    """Install the fish completion script."""
    path_obj = Path.home() / f".config/fish/completions/{prog_name}.fish"
    parent_dir: Path = path_obj.parent
    parent_dir.mkdir(parents=True, exist_ok=True)
    script_content = _fish_completion_script()
    path_obj.write_text(f"{script_content}\n")
    return path_obj


def install(
    *,
    prog_name: str,
    shell: str | None = None,
    complete_var: str | None = None,
) -> tuple[str, Path]:
    """Install the completion script for the given shell."""

    if complete_var is None:
        complete_var = "_{}_COMPLETE".format(prog_name.replace("-", "_").upper())
    if shell is None and shellingham is not None:
        shell, _ = shellingham.detect_shell()
        console_trace.print(f"[bright_black]Detected shell: {shell}")
    if shell == "bash":
        installed_path = install_bash(prog_name=prog_name)
        return shell, installed_path
    if shell == "zsh":
        installed_path = install_zsh(prog_name=prog_name)
        return shell, installed_path
    if shell == "fish":
        installed_path = install_fish(prog_name=prog_name)
        return shell, installed_path
    # No other shells supported at the moment
    err_console.print(f"[red] Shell {shell} is not supported.")
    sys.exit(1)


def install_subcommand_completions(subcommand: SubCommand):
    """Install the completion for a subcommand."""
    try:
        if isinstance(subcommand.tab_completion_install_option, str):
            args = [str(subcommand.executable), str(subcommand.tab_completion_install_option)]
        elif isinstance(subcommand.tab_completion_install_option, list):
            args = [str(subcommand.executable), *subcommand.tab_completion_install_option]
        else:
            raise TypeError(
                f"Invalid type for tab_completion_install_option: {type(subcommand.tab_completion_install_option)}"
            )

        install_completion_res = run(  # noqa: S603
            args, check=False, timeout=15
        )
        if install_completion_res.returncode != 0:
            err_console.print(f"[red]Failed to install completions for {subcommand.name}")
    except TimeoutExpired:
        err_console.print(f"[red]Failed to install completions for {subcommand.name}")


def install_completion():
    """Install the slices cli completion script."""
    # First, install completions for all subcommands that support it
    subcommands = find_subcommands()
    core_done = False
    for subcommand in subcommands:
        if subcommand.tab_completion_supported and subcommand.tab_completion_install_option:
            if subcommand.is_core:
                if core_done:
                    continue
                core_done = True
                console.print("[blue]Installing completion for 'slices-core'...")
                save_cached_subcommand_info(subcommand.executable, "core", "completion_installed", "true")
            else:
                console.print(f"[blue]Installing completion for 'slices-{subcommand.name}'...")
                save_cached_subcommand_info(subcommand.executable, subcommand.name, "completion_installed", "true")
            install_subcommand_completions(subcommand)
            console.print()
        if subcommand.tab_completion_supported and not subcommand.tab_completion_install_option:
            console.print(f"[red]Can not automatically install completions for 'slices-{subcommand.name}'.")
            console.print("[red bold]You need to install these manually.")
            console.print()

    # Finally, install the completion script for the slices executable itself
    console.print("[blue]Installing completion for 'slices'...")
    shell, path = install(prog_name="slices")
    console.print(f"[green]{shell} completion installed in {path}")
    console.print("Completion will take effect once you restart the terminal")
    sys.exit(0)


def print_completion_script():
    """Show the completion script for the given shell."""
    shell, _ = shellingham.detect_shell()
    console_trace.print(f"[bright_black]Detected shell: {shell}")
    if shell == "bash":
        print(_bash_completion_script())
        return
    if shell == "zsh":
        print(_zsh_completion_script())
        return
    if shell == "fish":
        print(_fish_completion_script())
        return
    # No other shells supported at the moment
    err_console.print(f"[red] Shell {shell} is not supported.")
    sys.exit(1)



"""Helpers for checking if a new CLI version is available."""

import json
from dataclasses import dataclass
from datetime import datetime, timedelta
from importlib.metadata import version as metadata_version

import requests
from packaging.version import Version
from pytz import UTC
from slices_cli_core import err_console

from . import cache_dir_location, console_trace


@dataclass
class AvailableVersionInfo:
    """Info about the available CLI version in the package repository."""

    stable_version: Version
    pre_version: Version | None
    info_updated_at: datetime


@dataclass
class InstalledVersionInfo:
    """Info about the locally installed CLI version."""

    version: Version

    @property
    def is_prerelease(self) -> bool:
        """Return True if a pre-release installed locally."""
        return self.version.is_prerelease


SLICES_CLI_PACKAGE_INFO_URL = "https://doc.slices-ri.eu/pypi/slices-cli/json"


def download_available_versions() -> AvailableVersionInfo:
    """Contact the package repository, to get info about the available CLI version."""
    response = requests.get(SLICES_CLI_PACKAGE_INFO_URL, timeout=5)
    response.raise_for_status()
    version_info = response.json()
    newest_stable_version = None
    newest_pre_version = None
    for version_str in version_info["releases"]:
        version = Version(version_str)
        if version.is_prerelease:
            if newest_pre_version is None or version > newest_pre_version:
                newest_pre_version = version
        elif newest_stable_version is None or version > newest_stable_version:
            newest_stable_version = version
    assert newest_stable_version is not None
    return AvailableVersionInfo(
        stable_version=newest_stable_version, pre_version=newest_pre_version, info_updated_at=datetime.now(UTC)
    )


def get_running_version() -> InstalledVersionInfo:
    """Return the running Slices CLI version info."""
    return InstalledVersionInfo(Version(metadata_version("slices-cli")))


version_cache_file_location = cache_dir_location / "cache_package_info.json"
version_warning_date_file_location = cache_dir_location / "cache_new_version_warning.json"


def save_version_cache(version_info: AvailableVersionInfo):
    """Save the AvailableVersionInfo to the cache."""
    try:
        with version_cache_file_location.open("w") as version_cache_file:
            json.dump(
                {
                    "stable_version": str(version_info.stable_version),
                    "pre_version": str(version_info.pre_version),
                    "info_updated_at": version_info.info_updated_at.timestamp(),
                },
                version_cache_file,
            )
    except:  # noqa: E722, S110
        # ignore
        # raise
        pass


def load_version_cache() -> AvailableVersionInfo | None:
    """Load the AvailableVersionInfo from the cache."""
    if not version_cache_file_location.exists():
        return None
    try:
        with version_cache_file_location.open() as version_cache_file:
            version_cache = json.load(version_cache_file)
            return AvailableVersionInfo(
                stable_version=Version(version_cache["stable_version"]),
                pre_version=Version(version_cache["pre_version"]),
                info_updated_at=datetime.fromtimestamp(version_cache["info_updated_at"], tz=UTC),
            )
    except:  # noqa: E722
        # ignore
        # raise
        return None


def save_version_warning_date(warned_at: datetime):
    """Save the version warning date."""
    try:
        with version_warning_date_file_location.open("w") as version_warning_date_file:
            json.dump({"warned_at": warned_at.timestamp()}, version_warning_date_file)
    except:  # noqa: E722, S110
        # ignore
        pass


def load_version_warning_date() -> datetime | None:
    """Load the version warning date."""
    if not version_warning_date_file_location.exists():
        return None
    try:
        with version_warning_date_file_location.open() as version_warning_date_file:
            version_warning_date = json.load(version_warning_date_file)
            return datetime.fromtimestamp(version_warning_date["warned_at"], tz=UTC)
    except:  # noqa: E722
        # ignore
        return None


UPGRADE_CHECK_INTERVAL = timedelta(hours=1)
UPGRADE_WARN_INTERVAL = timedelta(hours=20)


def is_newer_version_available() -> tuple[InstalledVersionInfo, AvailableVersionInfo] | None:
    """Return version info, or None if no newer version available."""

    available = load_version_cache()
    if available is None or available.info_updated_at < datetime.now(UTC) - UPGRADE_CHECK_INTERVAL:
        available = download_available_versions()
        if available is None:
            return None
        save_version_cache(available)
    installed = get_running_version()
    if installed.is_prerelease:
        if available.pre_version and installed.version < available.pre_version:
            return installed, available
    elif installed.version < available.stable_version:
        return installed, available
    return None


def do_upgrade_check():
    """Check if a new version is available, and output info to console if it is."""
    from rich.panel import Panel  # noqa: PLC0415  lazy loading of rich

    try:
        console_trace.print("[bright_black]Checking for new version...")
        has_newer = is_newer_version_available()
        if not has_newer:
            console_trace.print("[bright_black]   No new version available.")
            return
        warned_at = load_version_warning_date()
        if warned_at is not None and warned_at > datetime.now(UTC) - UPGRADE_WARN_INTERVAL:
            # Already warned recently
            console_trace.print("[bright_black]   New version available, but already warned recently.")
            return
        installed, available = has_newer
        new_v = available.pre_version if installed.is_prerelease else available.stable_version
        if installed.is_prerelease:
            command = "pip install --upgrade --pre slices-cli --extra-index-url=https://doc.slices-ri.eu/pypi/"
        else:
            command = "pip install --upgrade slices-cli --extra-index-url=https://doc.slices-ri.eu/pypi/"
        err_console.print(
            Panel(
                f"""Current version: {installed.version}
     New version: [bold]{new_v}[/bold]
    Upgrade with: [italic]{command}[/italic]""",
                title="[yellow]New Slices CLI version available",
                expand=False,
                style="yellow",
                padding=(1, 3),
            )
        )
        save_version_warning_date(datetime.now(UTC))
    except Exception as e:
        # If anything at all goes wrong, ignore and return no new version is available
        err_console.print(f"Problem checking for new version: {e}")
        return



"""Find version of executable."""

from pathlib import Path
from subprocess import TimeoutExpired, run


def get_executable_version(executable: Path) -> str:
    """Return the version of the executable."""

    # print(f"get_executable_version {executable}")
    try:
        version_res = run([str(executable), "--version"], check=False, capture_output=True, timeout=15)  # noqa: S603
        if version_res.returncode != 0:
            return "<version fetch error>"
        return version_res.stdout.decode("utf-8").strip().replace("\n", " ")
    except TimeoutExpired:
        return "<version fetch timeout>"



"""Functions for cache management."""

import json
from pathlib import Path

from . import cache_dir_location, config_dir_location, console, console_info


def clear_cache():
    """Delete all cache files."""

    for file in cache_dir_location.iterdir():
        # extra safety: only delete files ending with .cache  (old)
        #               and files starting with cache_ (new)
        if file.name.endswith(".cache") or file.name.startswith("cache_"):
            console_info.print(f"[bright_black]Deleting cache file: {file}")
            file.unlink()
    console.print("[green]:heavy_check_mark: Cache cleared successfully ")


def _cached_subcommand_manifest_filename(name: str) -> Path:
    return cache_dir_location / f"cache_subcommand_manifest_{name}.json"


def _cached_subcommand_info_filename(name: str) -> Path:
    return cache_dir_location / f"cache_subcommand_info_{name}.json"


def get_cached_subcommand_manifest(executable: Path, name: str) -> dict | None:
    """Return the cached manifest for the given subcommand, or None if not cached."""

    f = _cached_subcommand_manifest_filename(name)
    if f.is_file():
        current_exec_modified_time = int(executable.stat().st_mtime)
        with f.open() as cached_manifest_file:
            cached_manifest = json.load(cached_manifest_file)
            cached_exec_modified_time = cached_manifest["modified_time"]
            del cached_manifest["modified_time"]
        if cached_exec_modified_time == current_exec_modified_time:
            return cached_manifest
        # manifest is outdated, delete it
        f.unlink()
    return None


def save_cached_subcommand_manifest(executable: Path, name: str, manifest: dict):
    """Save the manifest for the given subcommand to the cache."""

    if not config_dir_location.is_dir():
        config_dir_location.mkdir()
    if not cache_dir_location.is_dir():
        cache_dir_location.mkdir()
    f = _cached_subcommand_manifest_filename(name)
    to_save = {**manifest, "modified_time": int(executable.stat().st_mtime)}
    with f.open("w") as cached_manifest_file:
        json.dump(to_save, cached_manifest_file)


def get_cached_subcommand_info(executable: Path, name: str, key: str, *, invalidate_on_exe_update: bool) -> str | None:
    """Return cached extra info for the given subcommand, or None if not cached."""

    f = _cached_subcommand_info_filename(name)
    if f.is_file():
        current_exec_modified_time = int(executable.stat().st_mtime)
        with f.open() as cached_info_file:
            cached_info = json.load(cached_info_file)
            if key not in cached_info:
                return None
            cached_entry = cached_info[key]
            if "modified_time" not in cached_entry or "value" not in cached_entry:
                return None
            cached_exec_modified_time = cached_entry["modified_time"]
        if cached_exec_modified_time == current_exec_modified_time or not invalidate_on_exe_update:
            return cached_entry["value"]
        # key is outdated. We ignore it.
    return None


def save_cached_subcommand_info(executable: Path, name: str, key: str, value: str | None):
    """Save an extra info entry for the given subcommand to the cache."""

    if not config_dir_location.is_dir():
        config_dir_location.mkdir()
    if not cache_dir_location.is_dir():
        cache_dir_location.mkdir()
    f = _cached_subcommand_info_filename(name)

    if f.is_file():
        with f.open() as cached_info_file:
            old_cached_info = json.load(cached_info_file)
    else:
        old_cached_info = {}

    to_save = {**old_cached_info, key: {"modified_time": int(executable.stat().st_mtime), "value": value}}
    with f.open("w") as cached_info_file:
        json.dump(to_save, cached_info_file)
